"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import sys
import warnings

from ... import mistune
from ...conversion import Converter


class Style:
    def __init__(self, display_name=None):
        if display_name:
            self.display_name = display_name
        else:
            self.display_name = ""

    def __repr__(self):
        s = ""
        for attribute in vars(self):
            if getattr(self, attribute) and attribute != "display_name":
                s += f"{self.display_name}.{attribute}: {getattr(self, attribute)}\n"
        return s.replace("\n\n", "\n")


class FontStyle(Style):
    def __init__(
        self,
        display_name=None,
        color=None,
        size=None,
        bold=None,
        italic=None,
        name=None,
    ):
        super().__init__(display_name=display_name)
        self.color = color
        self.size = size
        self.bold = bold
        self.italic = italic
        self.name = name


class MarkdownStyle:
    """
    ``MarkdownStyle`` defines how ``Markdown`` objects are being rendered in Excel cells
    or shapes. Start by instantiating a ``MarkdownStyle`` object. Printing it will show
    you the current (default) style:

    >>> style = MarkdownStyle()
    >>> style
    <MarkdownStyle>
    h1.font: .bold: True
    h1.blank_lines_after: 1
    paragraph.blank_lines_after: 1
    unordered_list.bullet_character: •
    unordered_list.blank_lines_after: 1
    strong.bold: True
    emphasis.italic: True

    You can override the defaults, e.g., to make ``**strong**`` text red instead of
    bold, do this:

    >>> style.strong.bold = False
    >>> style.strong.color = (255, 0, 0)
    >>> style.strong
    strong.color: (255, 0, 0)

    .. versionadded:: 0.23.0
    """

    class __Heading1(Style):
        def __init__(self):
            super().__init__(display_name="h1")
            self.font = FontStyle(bold=True)
            self.blank_lines_after = 0

    class __Paragraph(Style):
        def __init__(self):
            super().__init__(display_name="paragraph")
            self.blank_lines_after = 1

    class __UnorderedList(Style):
        def __init__(self):
            super().__init__(display_name="unordered_list")
            self.bullet_character = "\u2022"
            self.blank_lines_after = 1

    def __init__(self):
        self.h1 = self.__Heading1()
        self.paragraph = self.__Paragraph()
        self.unordered_list = self.__UnorderedList()
        self.strong = FontStyle(display_name="strong", bold=True)
        self.emphasis = FontStyle(display_name="emphasis", italic=True)

    def __repr__(self):
        s = "<MarkdownStyle>\n"
        for attribute in vars(self):
            s += f"{getattr(self, attribute)}"
        return s


class Markdown:
    """
    Markdown objects can be assigned to a single cell or shape via ``myrange.value`` or
    ``myshape.text``. They accept a string in Markdown format which will cause the text
    in the cell to be formatted accordingly. They can also be used in
    ``mysheet.render_template()``.

    .. note:: On macOS, formatting is currently not supported, but things like bullet
              points will still work.

    Arguments
    ---------
    text : str
        The text in Markdown syntax

    style : MarkdownStyle object, optional
        The MarkdownStyle object defines how the text will be formatted.

    Examples
    --------

    >>> mysheet['A1'].value = Markdown("A text with *emphasis* and **strong** style.")
    >>> myshape.text = Markdown("A text with *emphasis* and **strong** style.")

    .. versionadded:: 0.23.0
    """

    def __init__(self, text, style=MarkdownStyle()):
        self.text = text
        self.style = style


class MarkdownConverter(Converter):
    @classmethod
    def write_value(cls, value, options):
        return render_text(value.text, value.style)


def traverse_ast_node(tree, data=None, level=0):
    data = (
        {
            "length": [],
            "type": [],
            "parent_type": [],
            "text": [],
            "parents": [],
            "level": [],
        }
        if data is None
        else data
    )
    for element in tree:
        data["parents"] = data["parents"][:level]
        if "children" in element:
            data["parents"].append(element)
            traverse_ast_node(element["children"], data, level=level + 1)
        else:
            data["level"].append(level)
            data["parent_type"].append([parent["type"] for parent in data["parents"]])
            data["type"].append(element["type"])
            if element["type"] == "text":
                data["length"].append(len(element["text"]))
                data["text"].append(element["text"])
            elif element["type"] == "linebreak":
                data["length"].append(1)
                data["text"].append("\n")
    return data


def flatten_ast(value):
    parse_ast = mistune.create_markdown(renderer=mistune.AstRenderer())
    ast = parse_ast(value)
    flat_ast = []
    for node in ast:
        rv = traverse_ast_node([node])
        del rv["parents"]
        flat_ast.append(rv)
    return flat_ast


def render_text(text, style):
    flat_ast = flatten_ast(text)
    output = ""
    for node in flat_ast:
        # heading/list currently don't respect the level
        if "heading" in node["parent_type"][0]:
            output += "".join(node["text"])
            output += "\n" + style.h1.blank_lines_after * "\n"
        elif "paragraph" in node["parent_type"][0]:
            output += "".join(node["text"])
            output += "\n" + style.paragraph.blank_lines_after * "\n"
        elif "list" in node["parent_type"][0]:
            for j in node["text"]:
                output += f"{style.unordered_list.bullet_character} {j}\n"
            output += style.unordered_list.blank_lines_after * "\n"
    return output.rstrip("\n")


def format_text(parent, text, style):
    if sys.platform.startswith("darwin"):
        # Characters formatting is broken because of a bug in AppleScript/Excel 2016
        warnings.warn("Markdown formatting is currently ignored on macOS.")
        return
    flat_ast = flatten_ast(text)
    position = 0
    for node in flat_ast:
        if "heading" in node["parent_type"][0]:
            node_length = sum(node["length"]) + style.h1.blank_lines_after + 1
            apply_style_to_font(
                style.h1.font, parent.characters[position : position + node_length].font
            )
        elif "paragraph" in node["parent_type"][0]:
            node_length = sum(node["length"]) + style.paragraph.blank_lines_after + 1
            intra_node_position = position
            for ix, j in enumerate(node["parent_type"]):
                selection = slice(
                    intra_node_position, intra_node_position + node["length"][ix]
                )
                if "strong" in j:
                    apply_style_to_font(style.strong, parent.characters[selection].font)
                elif "emphasis" in j:
                    apply_style_to_font(
                        style.emphasis, parent.characters[selection].font
                    )
                intra_node_position += node["length"][ix]
        elif "list" in node["parent_type"][0]:
            node_length = sum(node["length"]) + style.unordered_list.blank_lines_after
            for _ in node["text"]:
                # TODO: check ast level to allow nested **strong** etc.
                node_length += 3  # bullet, space and new line
        else:
            node_length = sum(node["length"])
        position += node_length


def apply_style_to_font(style_object, font_object):
    for attribute in vars(style_object):
        if getattr(style_object, attribute):
            setattr(font_object, attribute, getattr(style_object, attribute))



# TODO: add all classes and use in _mac.py and _windows.py


class Apps:
    def keys(self):
        raise NotImplementedError()

    def add(self, spec=None, add_book=None, xl=None, visible=None):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __getitem__(self, pid):
        raise NotImplementedError()


class App:
    @property
    def xl(self):
        raise NotImplementedError()

    @xl.setter
    def xl(self, value):
        raise NotImplementedError()

    @property
    def api(self):
        raise NotImplementedError()

    @property
    def selection(self):
        raise NotImplementedError()

    def activate(self, steal_focus=False):
        raise NotImplementedError()

    @property
    def visible(self):
        raise NotImplementedError()

    @visible.setter
    def visible(self, visible):
        raise NotImplementedError()

    def quit(self):
        raise NotImplementedError()

    def kill(self):
        raise NotImplementedError()

    @property
    def screen_updating(self):
        raise NotImplementedError()

    @screen_updating.setter
    def screen_updating(self, value):
        raise NotImplementedError()

    @property
    def display_alerts(self):
        raise NotImplementedError()

    @display_alerts.setter
    def display_alerts(self, value):
        raise NotImplementedError()

    @property
    def enable_events(self):
        raise NotImplementedError()

    @enable_events.setter
    def enable_events(self, value):
        raise NotImplementedError()

    @property
    def interactive(self):
        raise NotImplementedError()

    @interactive.setter
    def interactive(self, value):
        raise NotImplementedError()

    @property
    def startup_path(self):
        raise NotImplementedError()

    @property
    def calculation(self):
        raise NotImplementedError()

    @calculation.setter
    def calculation(self, value):
        raise NotImplementedError()

    def calculate(self):
        raise NotImplementedError()

    @property
    def version(self):
        raise NotImplementedError()

    @property
    def books(self):
        raise NotImplementedError()

    @property
    def hwnd(self):
        raise NotImplementedError()

    @property
    def pid(self):
        raise NotImplementedError()

    def run(self, macro, args):
        raise NotImplementedError()

    @property
    def status_bar(self):
        raise NotImplementedError()

    @status_bar.setter
    def status_bar(self, value):
        raise NotImplementedError()

    @property
    def cut_copy_mode(self):
        raise NotImplementedError()

    @cut_copy_mode.setter
    def cut_copy_mode(self, value):
        raise NotImplementedError()

    def alert(self, prompt, title, buttons, mode, callback):
        raise NotImplementedError()


class Books:
    @property
    def api(self):
        raise NotImplementedError()

    @property
    def active(self):
        raise NotImplementedError()

    def __call__(self, name_or_index):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def add(self):
        raise NotImplementedError()

    def open(
        self,
        fullname,
        update_links=None,
        read_only=None,
        format=None,
        password=None,
        write_res_password=None,
        ignore_read_only_recommended=None,
        origin=None,
        delimiter=None,
        editable=None,
        notify=None,
        converter=None,
        add_to_mru=None,
        local=None,
        corrupt_load=None,
    ):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()


class Book:
    @property
    def api(self):
        raise NotImplementedError()

    def json(self):
        raise NotImplementedError()

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def sheets(self):
        raise NotImplementedError()

    @property
    def app(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def save(self, path=None, password=None):
        raise NotImplementedError()

    @property
    def fullname(self):
        raise NotImplementedError()

    @property
    def names(self):
        raise NotImplementedError()

    def activate(self):
        raise NotImplementedError()

    def to_pdf(self, path, quality):
        raise NotImplementedError()


class Sheets:
    @property
    def api(self):
        raise NotImplementedError()

    @property
    def active(self):
        raise NotImplementedError()

    def __call__(self, name_or_index):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def add(self, before=None, after=None):
        raise NotImplementedError()


class Sheet:
    @property
    def api(self):
        raise NotImplementedError()

    @property
    def name(self):
        raise NotImplementedError()

    @name.setter
    def name(self, value):
        raise NotImplementedError()

    @property
    def names(self):
        raise NotImplementedError()

    @property
    def book(self):
        raise NotImplementedError()

    @property
    def index(self):
        raise NotImplementedError()

    def range(self, arg1, arg2=None):
        raise NotImplementedError()

    @property
    def cells(self):
        raise NotImplementedError()

    def activate(self):
        raise NotImplementedError()

    def select(self):
        raise NotImplementedError()

    def clear_contents(self):
        raise NotImplementedError()

    def clear_formats(self):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()

    def autofit(self, axis=None):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    def copy(self, before, after):
        raise NotImplementedError()

    @property
    def charts(self):
        raise NotImplementedError()

    @property
    def shapes(self):
        raise NotImplementedError()

    @property
    def tables(self):
        raise NotImplementedError()

    @property
    def pictures(self):
        raise NotImplementedError()

    @property
    def used_range(self):
        raise NotImplementedError()

    @property
    def visible(self):
        raise NotImplementedError()

    @visible.setter
    def visible(self, value):
        raise NotImplementedError()

    @property
    def page_setup(self):
        raise NotImplementedError()

    def to_pdf(self, path, quality):
        raise NotImplementedError()


class Range:
    @property
    def coords(self):
        raise NotImplementedError()

    @property
    def api(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    @property
    def row(self):
        raise NotImplementedError()

    @property
    def column(self):
        raise NotImplementedError()

    @property
    def shape(self):
        raise NotImplementedError()

    @property
    def raw_value(self):
        raise NotImplementedError()

    @raw_value.setter
    def raw_value(self, value):
        raise NotImplementedError()

    def clear_contents(self):
        raise NotImplementedError()

    def clear_formats(self):
        raise NotImplementedError()

    def clear(self):
        raise NotImplementedError()

    def end(self, direction):
        raise NotImplementedError()

    @property
    def formula(self):
        raise NotImplementedError()

    @formula.setter
    def formula(self, value):
        raise NotImplementedError()

    @property
    def formula2(self):
        raise NotImplementedError()

    @formula2.setter
    def formula2(self, value):
        raise NotImplementedError()

    @property
    def formula_array(self):
        raise NotImplementedError()

    @formula_array.setter
    def formula_array(self, value):
        raise NotImplementedError()

    @property
    def font(self):
        raise NotImplementedError()

    @property
    def column_width(self):
        raise NotImplementedError()

    @column_width.setter
    def column_width(self, value):
        raise NotImplementedError()

    @property
    def row_height(self):
        raise NotImplementedError()

    @row_height.setter
    def row_height(self, value):
        raise NotImplementedError()

    @property
    def width(self):
        raise NotImplementedError()

    @property
    def height(self):
        raise NotImplementedError()

    @property
    def left(self):
        raise NotImplementedError()

    @property
    def top(self):
        raise NotImplementedError()

    @property
    def has_array(self):
        raise NotImplementedError()

    @property
    def number_format(self):
        raise NotImplementedError()

    @number_format.setter
    def number_format(self, value):
        raise NotImplementedError()

    def get_address(self, row_absolute, col_absolute, external):
        raise NotImplementedError()

    @property
    def address(self):
        raise NotImplementedError()

    @property
    def current_region(self):
        raise NotImplementedError()

    def autofit(self, axis=None):
        raise NotImplementedError()

    def insert(self, shift=None, copy_origin=None):
        raise NotImplementedError()

    def delete(self, shift=None):
        raise NotImplementedError()

    def copy(self, destination=None):
        raise NotImplementedError()

    def paste(self, paste=None, operation=None, skip_blanks=False, transpose=False):
        raise NotImplementedError()

    @property
    def hyperlink(self):
        raise NotImplementedError()

    def add_hyperlink(self, address, text_to_display=None, screen_tip=None):
        raise NotImplementedError()

    @property
    def color(self):
        raise NotImplementedError()

    @color.setter
    def color(self, color_or_rgb):
        raise NotImplementedError()

    @property
    def name(self):
        raise NotImplementedError()

    @name.setter
    def name(self, value):
        raise NotImplementedError()

    def __call__(self, arg1, arg2=None):
        raise NotImplementedError()

    @property
    def rows(self):
        raise NotImplementedError()

    @property
    def columns(self):
        raise NotImplementedError()

    def select(self):
        raise NotImplementedError()

    @property
    def merge_area(self):
        raise NotImplementedError()

    @property
    def merge_cells(self):
        raise NotImplementedError()

    def merge(self, across):
        raise NotImplementedError()

    def unmerge(self):
        raise NotImplementedError()

    @property
    def table(self):
        raise NotImplementedError()

    @property
    def characters(self):
        raise NotImplementedError()

    @property
    def wrap_text(self):
        raise NotImplementedError()

    @wrap_text.setter
    def wrap_text(self, value):
        raise NotImplementedError()

    @property
    def note(self):
        raise NotImplementedError()

    def copy_picture(self, appearance, format):
        raise NotImplementedError()

    def to_png(self, path):
        raise NotImplementedError()

    def to_pdf(self, path, quality):
        raise NotImplementedError()


class Picture:
    @property
    def api(self):
        raise NotImplementedError()

    @property
    def name(self):
        raise NotImplementedError()

    @name.setter
    def name(self, value):
        raise NotImplementedError()

    @property
    def parent(self):
        raise NotImplementedError()

    @property
    def left(self):
        raise NotImplementedError()

    @left.setter
    def left(self, value):
        raise NotImplementedError()

    @property
    def top(self):
        raise NotImplementedError()

    @top.setter
    def top(self, value):
        raise NotImplementedError()

    @property
    def width(self):
        raise NotImplementedError()

    @width.setter
    def width(self, value):
        raise NotImplementedError()

    @property
    def height(self):
        raise NotImplementedError()

    @height.setter
    def height(self, value):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()

    @property
    def lock_aspect_ratio(self):
        raise NotImplementedError()

    @lock_aspect_ratio.setter
    def lock_aspect_ratio(self, value):
        raise NotImplementedError()

    def index(self):
        raise NotImplementedError()


class Collection:
    @property
    def api(self):
        raise NotImplementedError()

    @property
    def parent(self):
        raise NotImplementedError()

    def __call__(self, key):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def __contains__(self, key):
        raise NotImplementedError()


class Pictures:
    def add(self, filename, link_to_file, save_with_document, left, top, width, height):
        raise NotImplementedError()



import re
from ..util import escape_html


DEF_ABBR = re.compile(
    # *[HTML]:
    # *[HTML]: Hyper Text Markup Language
    # *[HTML]:
    #     Hyper Text Markup Language
    r'\*\[([^\]]+)\]:'
    r'((?:[ \t]*\n(?: {3,}|\t)[^\n]+)|(?:[^\n]*))\n*'
)


def parse_def_abbr(block, m, state):
    def_abbrs = state.get('def_abbrs', {})
    label = m.group(1)
    definition = m.group(2)
    def_abbrs[label] = definition.strip()
    state['def_abbrs'] = def_abbrs


def parse_inline_abbr(inline, m, state):
    def_abbrs = state['def_abbrs']
    label = m.group(0)
    return 'abbr', label, def_abbrs[label]


def after_parse_def_abbr(md, tokens, state):
    def_abbrs = state.get('def_abbrs')
    if def_abbrs:
        labels = list(def_abbrs.keys())
        abbr_pattern = r'|'.join(re.escape(k) for k in labels)
        md.inline.register_rule('abbr', abbr_pattern, parse_inline_abbr)
        md.inline.rules.append('abbr')
    return tokens


def render_html_abbr(key, definition):
    title_attribute = ""
    if definition:
        definition = escape_html(definition)
        title_attribute = ' title="{}"'.format(definition)

    return "<abbr{title_attribute}>{key}</abbr>".format(
        key=key,
        title_attribute=title_attribute,
    )


def render_ast_abbr(key, definition):
    return {'type': 'abbr', 'text': key, 'definition': definition}


def plugin_abbr(md):
    md.block.register_rule('def_abbr', DEF_ABBR, parse_def_abbr)
    md.before_render_hooks.append(after_parse_def_abbr)
    md.block.rules.append('def_abbr')

    if md.renderer.NAME == 'html':
        md.renderer.register('abbr', render_html_abbr)
    elif md.renderer.NAME == 'ast':
        md.renderer.register('abbr', render_ast_abbr)



try:
    import pandas as pd
except ImportError:
    pd = None


if pd:
    import numpy as np
    from . import Converter, Options

    def write_value(cls, value, options):
        index = options.get("index", True)
        header = options.get("header", True)
        assign_empty_index_names = options.get("assign_empty_index_names", False)

        index_names = value.index.names
        if assign_empty_index_names:
            # Useful when you want to have your DataFrame formatted as an Excel table
            # which requires column header names. Since Excel tables only allow an empty
            # space once, we'll generate multiple empty spaces for each column.
            index_names = [
                f" " * (i + 1) if name is None else name
                for i, name in enumerate(index_names)
            ]
        else:
            index_names = ["" if name is None else name for name in index_names]
        index_levels = len(index_names)

        if index:
            if value.index.name in value.columns:
                # Prevents column name collision when resetting the index
                value.index.rename(None, inplace=True)
            value = value.reset_index()

        # Convert pandas-specific types without corresponding Excel type to strings
        for ix, col in enumerate(value.columns):
            if (
                isinstance(value.iloc[:, ix].dtype, pd.PeriodDtype)
                or value.iloc[:, ix].dtype == "timedelta64[ns]"
            ):
                value.iloc[:, ix] = value.iloc[:, ix].astype(str)

        if header:
            if isinstance(value.columns, pd.MultiIndex):
                columns = list(zip(*value.columns.tolist()))
                columns = [list(i) for i in columns]
                # Move index names right above the index
                if index:
                    for c in columns[:-1]:
                        c[:index_levels] = [""] * index_levels
                    columns[-1][:index_levels] = index_names
            else:
                columns = [value.columns.tolist()]
                if index:
                    columns[0][:index_levels] = index_names
            value = columns + value.values.tolist()
        else:
            value = value.values.tolist()

        return value

    class PandasDataFrameConverter(Converter):

        writes_types = pd.DataFrame

        @classmethod
        def base_reader(cls, options):
            return super(PandasDataFrameConverter, cls).base_reader(
                Options(options).override(ndim=2)
            )

        @classmethod
        def read_value(cls, value, options):
            index = options.get("index", 1)
            header = options.get("header", 1)
            dtype = options.get("dtype", None)
            copy = options.get("copy", False)

            # build dataframe with only columns (no index) but correct header
            if header == 1:
                columns = pd.Index(value[0])
            elif header > 1:
                columns = pd.MultiIndex.from_arrays(value[:header])
            else:
                columns = None

            df = pd.DataFrame(value[header:], columns=columns, dtype=dtype, copy=copy)

            # handle index by resetting the index to the index first columns
            # and renaming the index according to the name in the last row
            if index > 0:
                # rename uniquely the index columns to some never used name for column
                # we do not use the column name directly as it would cause issues if
                # several columns have the same name
                df.columns = pd.Index(range(len(df.columns)))
                df.set_index(list(df.columns)[:index], inplace=True)

                df.index.names = pd.Index(
                    value[header - 1][:index] if header else [None] * index
                )

                if header:
                    df.columns = columns[index:]
                else:
                    df.columns = pd.Index(range(len(df.columns)))

            return df

        @classmethod
        def write_value(cls, value, options):
            return write_value(cls, value, options)

    PandasDataFrameConverter.register(pd.DataFrame, "df")

    class PandasSeriesConverter(Converter):

        writes_types = pd.Series

        @classmethod
        def read_value(cls, value, options):
            index = options.get("index", 1)
            header = options.get("header", True)
            dtype = options.get("dtype", None)
            copy = options.get("copy", False)

            if header:
                columns = value[0]
                if not isinstance(columns, list):
                    columns = [columns]
                data = value[1:]
            else:
                columns = None
                data = value

            df = pd.DataFrame(data, columns=columns, dtype=dtype, copy=copy)

            if index:
                df.columns = pd.Index(range(len(df.columns)))
                df.set_index(list(df.columns)[:index], inplace=True)
                df.index.names = pd.Index(
                    value[header - 1][:index] if header else [None] * index
                )

            if header:
                df.columns = columns[index:]
            else:
                df.columns = pd.Index(range(len(df.columns)))

            series = df.squeeze()

            if not header:
                series.name = None
                series.index.name = None

            return series

        @classmethod
        def write_value(cls, value, options):
            if all(v is None for v in value.index.names) and value.name is None:
                default_header = False
            else:
                default_header = True

            options["header"] = options.get("header", default_header)
            values = write_value(cls, value.to_frame(), options)
            return values

    PandasSeriesConverter.register(pd.Series)



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import sys
import shutil
import datetime as dt
import numbers
import warnings

try:
    from jinja2 import Environment, nodes
except ImportError:
    Environment, nodes = None, None

from . import filters
from .markdown import Markdown
from .image import Image
from ..utils import LicenseHandler
from ...main import Book, XlwingsError

try:
    import PIL
    import PIL.Image
except ImportError:
    PIL = None

try:
    from matplotlib.figure import Figure
except ImportError:
    Figure = None

try:
    import numpy as np
except ImportError:
    np = None

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import plotly
except ImportError:
    plotly = None

LicenseHandler.validate_license("reports")


def parse_single_placeholder(value, env):
    """This is only for cells that contain a single placeholder.
    Text with multiple placeholders is handled by Jinja's native (custom) filter system.
    Returns var, filter_list with filter_name:filter_args (list of dicts)
    """
    ast = env.parse(value)
    found_nodes = list(ast.find_all(node_type=nodes.Filter))
    if found_nodes:
        node = found_nodes[0]
        f = node
        filter_list = [{f.name: f.args}]
        while isinstance(f.node, nodes.Filter):
            f = f.node
            filter_list.insert(0, {f.name: f.args})
        return f.node.name, filter_list
    else:
        return value.replace("{{", "").replace("}}", "").strip(), []


def render_sheet(sheet, **data):
    """
    Replaces the Jinja2 placeholders in a given sheet
    """
    if sheet.name.startswith("##"):
        return
    book = sheet.book

    # Shapes aren't properly moved otherwise
    sheet.select()

    # Inserting rows with Frames changes the print area.
    # Get it here so we can revert at the end.
    print_area = sheet.page_setup.print_area

    # A Jinja env defines the placeholder markers and allows to register custom filters
    if not Environment:
        raise XlwingsError(
            "Couldn't find the 'jinja2' package, which is required for xlwings Reports."
        )
    env = Environment()
    env.filters["datetime"] = filters.datetime
    env.filters["format"] = filters.fmt  # This overrides Jinja's built-in filter
    env.filters["fontcolor"] = filters.fontcolor

    # used_range doesn't start automatically in A1
    last_cell = sheet.used_range.last_cell
    values_all = (
        sheet.range((1, 1), (last_cell.row, last_cell.column)).options(ndim=2).value
        if sheet.used_range.value
        else []
    )

    # Frames
    uses_frames = False
    frame_indices = []
    for ix, cell in enumerate(sheet.range((1, 1), (1, last_cell.column))):
        if cell.note:
            if "<frame>" in cell.note.text:
                frame_indices.append(ix)
                uses_frames = True
    is_single_frame = True if len(frame_indices) == 1 else False
    frame_indices += [0, last_cell.column]
    frame_indices = list(sorted(set(frame_indices)))
    values_per_frame = []
    for ix in range(len(frame_indices) - 1):
        values_per_frame.append(
            [i[frame_indices[ix] : frame_indices[ix + 1]] for i in values_all]
        )

    # Loop through every cell for each frame
    for ix, values in enumerate(values_per_frame):
        row_shift = 0
        for i, row in enumerate(values):
            for j, value in enumerate(row):
                cell = sheet[i + row_shift, j + frame_indices[ix]]
                if isinstance(value, str):
                    if (
                        value.count("{{") == 1
                        and value.startswith("{{")
                        and value.endswith("}}")
                    ):
                        # Cell contains single Jinja variable
                        var, filter_list = parse_single_placeholder(value, env)
                        result = env.compile_expression(var)(**data)
                        if (
                            isinstance(result, Image)
                            or (PIL and isinstance(result, PIL.Image.Image))
                            or (Figure and isinstance(result, Figure))
                            or (plotly and isinstance(result, plotly.graph_objs.Figure))
                        ):

                            # Image filters: these filters can only be used once. If
                            # supplied multiple times, the first one will be used.
                            width = filters.width(filter_list)
                            height = filters.height(filter_list)
                            scale = filters.scale(filter_list)
                            format_ = filters.image_format(filter_list)
                            top = filters.top(filter_list)
                            left = filters.left(filter_list)

                            image_types = (Image, PIL.Image.Image) if PIL else (Image,)
                            image = (
                                result.filename
                                if isinstance(result, image_types)
                                else result
                            )
                            sheet.pictures.add(
                                image,
                                top=top + cell.top,
                                left=left + cell.left,
                                width=width,
                                height=height,
                                scale=scale,
                                format=format_,
                            )
                            cell.value = None
                        elif isinstance(result, (str, numbers.Number)):
                            if any(["fontcolor" in f for f in filter_list]):
                                cell.font.color = filters.fontcolor(
                                    filter_list=filter_list
                                )
                            cell.value = env.from_string(value).render(**data)
                        elif isinstance(result, Markdown):
                            # This will conveniently render placeholders
                            # within Markdown instances
                            cell.value = Markdown(
                                text=env.from_string(result.text).render(**data),
                                style=result.style,
                            )
                        elif isinstance(result, dt.datetime):
                            cell.value = env.from_string(value).render(**data)
                        else:
                            # Arrays
                            options = {"index": False, "header": True}  # defaults
                            if isinstance(result, (list, tuple)) and isinstance(
                                result[0], (list, tuple)
                            ):
                                result_len = len(result)
                            elif np and isinstance(result, np.ndarray):
                                result_len = len(result)
                            elif pd and isinstance(result, pd.DataFrame):
                                result = (
                                    result.copy()
                                )  # prevents manipulation of the df in the data dict
                                # DataFrame Filters
                                for filter_item in filter_list:
                                    for filter_name, filter_args in filter_item.items():
                                        if filter_name in (
                                            "showindex",
                                            "noheader",
                                            "vmerge",
                                        ):
                                            continue  # handled below
                                        func = getattr(filters, filter_name)
                                        result = func(result, filter_args)
                                # showindex is undocumented
                                # as df.reset_index() is preferred
                                options = {
                                    "index": any(
                                        ["showindex" in f for f in filter_list]
                                    ),
                                    "header": not any(
                                        ["noheader" in f for f in filter_list]
                                    ),
                                }

                                # Assumes 1 header row,
                                # MultiIndex headers aren't supported
                                if any(["header" in f for f in filter_list]):
                                    # Hack for the 'header' filter
                                    result_len = 1
                                else:
                                    result_len = (
                                        len(result) + 1
                                        if options["header"]
                                        else len(result)
                                    )
                            else:
                                result_len = 1
                            # Insert rows if within <frame>
                            # and 'result' is multiple rows high
                            rows_to_be_inserted = 0
                            if uses_frames and result_len > 1:
                                # Deduct header and first data row
                                rows_to_be_inserted = result_len - (
                                    2 if options["header"] else 1
                                )
                                if rows_to_be_inserted > 0:
                                    properties = (
                                        {"screen_updating": True}
                                        if sys.platform.startswith("win")
                                        else {}
                                    )
                                    with book.app.properties(**properties):
                                        # Windows doesn't move objects properly with
                                        # screen_updating=False. Since CopyOrigin is
                                        # not supported on Mac, we start copying two
                                        # rows below the header so the data row
                                        # formatting gets carried over.
                                        start_row = (
                                            i
                                            + row_shift
                                            + (3 if options["header"] else 2)
                                        )
                                        start_col = j + frame_indices[ix] + 1
                                        end_row = (
                                            i
                                            + row_shift
                                            + rows_to_be_inserted
                                            + (2 if options["header"] else 1)
                                        )
                                        end_col = frame_indices[ix] + len(values[0])
                                        if is_single_frame:
                                            # This will preserve the row height of rows
                                            # below the inserted ones
                                            sheet.range(
                                                f"{start_row}:{end_row}"
                                            ).insert("down")
                                        else:
                                            sheet.range(
                                                (start_row, start_col),
                                                (end_row, end_col),
                                            ).insert("down")
                                        # Inserting does not take over borders
                                        sheet.range(
                                            (start_row - 1, start_col),
                                            (start_row - 1, end_col),
                                        ).copy()
                                        sheet.range(
                                            (start_row - 1, start_col),
                                            (end_row, end_col),
                                        ).paste(paste="formats")
                            # Write the array to Excel
                            if cell.table:
                                cell.table.update(result, index=options["index"])
                            else:
                                cell.options(**options).value = result
                            # DataFrame formatting filters
                            for filter_item in filter_list:
                                for filter_name, filter_args in filter_item.items():
                                    if filter_name == "vmerge":
                                        merge_ranges = filters.vmerge(
                                            result, filter_args, cell, options["header"]
                                        )
                                        for merge_range in merge_ranges:
                                            merge_range.merge()
                            row_shift += rows_to_be_inserted
                    elif "{{" in value:
                        # These are strings with (multiple) Jinja variables so apply
                        # standard text rendering here
                        template = env.from_string(value)
                        cell.value = template.render(data)
                    else:
                        # Don't do anything with cells that don't contain any templating
                        # so we don't lose the formatting
                        pass

    # Loop through all shapes of interest with a template text
    for shape in [
        shape for shape in sheet.shapes if shape.type in ("auto_shape", "text_box")
    ]:
        shapetext = shape.text
        if shapetext and "{{" in shapetext:
            if (
                shapetext.count("{{") == 1
                and shapetext.startswith("{{")
                and shapetext.endswith("}}")
            ):
                # Single Jinja variable case, the only case we support with Markdown
                var, filter_list = parse_single_placeholder(shapetext, env)
                result = env.compile_expression(var)(**data)
                if isinstance(result, Markdown):
                    # This will conveniently render placeholders within Markdown text
                    shape.text = Markdown(
                        text=env.from_string(result.text).render(**data),
                        style=result.style,
                    )
                else:
                    # Single Jinja var but no Markdown
                    if any(["fontcolor" in f for f in filter_list]):
                        shape.font.color = filters.fontcolor(filter_list=filter_list)
                    template = env.from_string(shapetext)
                    shape.text = template.render(data)
            else:
                # Multiple Jinja vars and no Markdown
                template = env.from_string(shapetext)
                shape.text = template.render(data)

    # Copy/pasting the formatting leaves ranges selected.
    book.app.cut_copy_mode = False

    # Reset print area
    if print_area:
        sheet.page_setup.print_area = print_area

    try:
        sheet["A1"].select()
    except:
        pass


def create_report(template=None, output=None, book_settings=None, **data):
    warnings.warn("Deprecated. Use render_template instead.")
    return render_template(
        template=template, output=output, book_settings=book_settings, **data
    )


def render_template(template, output, book_settings=None, app=None, **data):
    """
    This function requires xlwings :guilabel:`PRO`.

    This is a convenience wrapper around
    :meth:`mysheet.render_template <xlwings.Sheet.render_template>`

    Writes the values of all key word arguments to the ``output`` file according to the
    ``template`` and the variables contained in there (Jinja variable syntax).
    Following variable types are supported:

    strings, numbers, lists, simple dicts, NumPy arrays, Pandas DataFrames, pictures and
    Matplotlib/Plotly figures.

    Parameters
    ----------
    template: str
        Path to your Excel template, e.g. ``r'C:\\Path\\to\\my_template.xlsx'``

    output: str
        Path to your Report, e.g. ``r'C:\\Path\\to\\my_report.xlsx'``

    book_settings: dict, default None
        A dict of ``xlwings.Book`` parameters, for details see: :attr:`xlwings.Book`.
        For example: ``book_settings={'update_links': False}``.

    app: xlwings App, default None
        By passing in an xlwings App instance, you can control where your report runs
        and configure things like ``visible=False``. For details see
        :attr:`xlwings.App`. By default, it creates the report in the currently active
        instance of Excel.

    data: kwargs
        All key/value pairs that are used in the template.

    Returns
    -------
    xlwings Book


    Examples
    --------
    In ``my_template.xlsx``, put the following Jinja variables in two cells:
    ``{{ title }}`` and ``{{ df }}``

    >>> from xlwings.pro.reports import render_template
    >>> import pandas as pd
    >>> df = pd.DataFrame(data=[[1,2],[3,4]])
    >>> mybook = render_template('my_template.xlsx', 'my_report.xlsx',
                                 title='MyTitle', df=df)

    With many template variables it may be useful to collect the data first:

    >>> data = dict(title='MyTitle', df=df)
    >>> mybook = render_template('my_template.xlsx', 'my_report.xlsx', **data)

    If you need to handle external links or a password, use it like so:

    >>> mybook = render_template('my_template.xlsx', 'my_report.xlsx',
                                 book_settings={'update_links': True,
                                 'password': 'mypassword'}, **data)
    """
    shutil.copyfile(template, output)
    if app:
        if book_settings:
            wb = app.books.open(output, **book_settings)
        else:
            wb = app.books.open(output)
    else:
        # Use existing Excel instance or create a new one if there is none
        if book_settings:
            wb = Book(output, **book_settings)
        else:
            wb = Book(output)

    for sheet in reversed(wb.sheets):
        render_sheet(sheet, **data)

    wb.save()
    return wb



"""
Required Notice: Copyright (C) Zoomer Analytics GmbH.

xlwings PRO is dual-licensed under one of the following licenses:

* PolyForm Noncommercial License 1.0.0 (for noncommercial use):
  https://polyformproject.org/licenses/noncommercial/1.0.0
* xlwings PRO License (for commercial use):
  https://github.com/xlwings/xlwings/blob/main/LICENSE_PRO.txt

Commercial licenses can be purchased at https://www.xlwings.org
"""

import os
import base64
import hmac
import hashlib
import json
import binascii
import datetime as dt
import warnings
import xlwings  # To prevent circular imports
import time
import glob
import shutil
import tempfile
import atexit
from functools import lru_cache

from ..utils import read_config_sheet

try:
    from cryptography.fernet import Fernet, InvalidToken
except ImportError as e:
    Fernet = None
    InvalidToken = None


class LicenseHandler:
    @staticmethod
    def get_cipher():
        try:
            return Fernet('kS2PJbJuXbNXC-sY9XFGg01OmEKuCnKAH0zyMNWWm8c=')
        except (TypeError, ValueError):
            raise xlwings.LicenseError("Couldn't validate license key.") from None

    @staticmethod
    def get_license():
        # Sheet config (only used by RunPython, UDFs use env var)
        try:
            sheet_license_key = read_config_sheet(xlwings.Book.caller()).get(
                "LICENSE_KEY"
            )
            if sheet_license_key:
                return sheet_license_key
        except:
            pass
        # User config file
        config_file = xlwings.USER_CONFIG_FILE
        if os.path.exists(config_file):
            with open(config_file, "r") as f:
                config = f.readlines()
            key = None
            for line in config:
                if line.split(",")[0] == '"LICENSE_KEY"':
                    key = line.split(",")[1].strip()[1:-1]
            if key:
                return key
        # Env Var - also used if LICENSE_KEY is in config sheet and called via UDF
        if os.getenv("XLWINGS_LICENSE_KEY"):
            return os.environ["XLWINGS_LICENSE_KEY"]
        raise xlwings.LicenseError("Couldn't find a license key.")

    @staticmethod
    @lru_cache()
    def validate_license(product, license_type=None):
        key = LicenseHandler.get_license()
        if key == "noncommercial":
            return {"license_type": "noncommercial"}
        if key.startswith("gA") and not Fernet:
            # Legacy up to 0.27.12
            raise ImportError(
                "You are using a legacy license key that requires the 'cryptography' "
                "package. Alternatively, contact us for a new license key."
            ) from None
        elif key.startswith("gA"):
            cipher_suite = LicenseHandler.get_cipher()
            try:
                license_info = json.loads(cipher_suite.decrypt(key.encode()).decode())
            except (binascii.Error, InvalidToken):
                raise xlwings.LicenseError("Invalid license key.") from None
        else:
            signature = hmac.new(
                'kS2PJbJuXbNXC-sY9XFGg01OmEKuCnKAH0zyMNWWm8c='.encode(),
                key[:-5].encode(),
                hashlib.sha256,
            ).hexdigest()
            if signature[:5] != key[-5:]:
                raise xlwings.LicenseError("Invalid license key.") from None
            else:
                try:
                    license_info = json.loads(
                        base64.urlsafe_b64decode(key[:-5]).decode()
                    )
                except:
                    raise xlwings.LicenseError("Invalid license key.") from None
        try:
            if (
                license_type == "developer"
                and license_info["license_type"] != "developer"
            ):
                raise xlwings.LicenseError(
                    "You need a developer license for this action."
                )
        except KeyError:
            raise xlwings.LicenseError(
                "You need a developer license for this action."
            ) from None
        if (
            "valid_until" not in license_info.keys()
            or "products" not in license_info.keys()
        ):
            raise xlwings.LicenseError("Invalid license key format.") from None
        license_valid_until = dt.datetime.strptime(
            license_info["valid_until"], "%Y-%m-%d"
        ).date()
        if dt.date.today() > license_valid_until:
            raise xlwings.LicenseError(
                "Your license expired on {}.".format(
                    license_valid_until.strftime("%Y-%m-%d")
                )
            ) from None
        if product not in license_info["products"]:
            raise xlwings.LicenseError(
                f"License key isn't valid for the '{product}' functionality."
            ) from None
        if (
            "version" in license_info.keys()
            and license_info["version"] != xlwings.__version__
        ):
            raise xlwings.LicenseError(
                "Your license key is only valid for xlwings v{0}".format(
                    license_info["version"]
                )
            ) from None
        if (license_valid_until - dt.date.today()) < dt.timedelta(days=30):
            warnings.warn(
                f"Your license key expires in "
                f"{(license_valid_until - dt.date.today()).days} days."
            )
        return license_info

    @staticmethod
    def create_deploy_key():
        license_info = LicenseHandler.validate_license("pro", license_type="developer")
        if license_info["license_type"] == "noncommercial":
            return "noncommercial"
        license_dict = json.dumps(
            {
                "version": xlwings.__version__,
                "products": license_info["products"],
                "valid_until": "2999-12-31",
                "license_type": "deploy_key",
            }
        ).encode()
        if LicenseHandler.get_license().startswith("gA"):
            # Legacy
            cipher_suite = LicenseHandler.get_cipher()
            return cipher_suite.encrypt(license_dict).decode()
        else:
            body = base64.urlsafe_b64encode(license_dict)
            signature = hmac.new(
                'kS2PJbJuXbNXC-sY9XFGg01OmEKuCnKAH0zyMNWWm8c='.encode(), body, hashlib.sha256
            ).hexdigest()
            return f"{body.decode()}{signature[:5]}"


def get_embedded_code_temp_dir():
    tmp_base_path = os.path.join(tempfile.gettempdir(), "xlwings")
    os.makedirs(tmp_base_path, exist_ok=True)
    try:
        # HACK: Clean up directories that are older than 30 days
        # This should be done in the C++ part when the Python process is killed
        for subdir in glob.glob(tmp_base_path + "/*/"):
            if os.path.getmtime(subdir) < time.time() - 30 * 86400:
                shutil.rmtree(subdir, ignore_errors=True)
    except Exception:
        pass  # we don't care if it fails
    tempdir = tempfile.mkdtemp(dir=tmp_base_path)
    # This only works for RunPython calls running outside the COM server
    atexit.register(shutil.rmtree, tempdir)
    return tempdir
